from ukpostcodes.fix import fix
from ukpostcodes.ukpostcode import parse_from_corpus, Postcode

#TODO - covert to actual tests

def test_fix():
    # Trims postcode
    assert fix(" SW1A 2AA ") == "SW1A 2AA"
    # Upper case
    assert fix(" Sw1A 2aa ") == "SW1A 2AA"
    # Fixes spacing
    assert fix(" Sw1A2aa ") == "SW1A 2AA"
    assert fix(" Sw1A    2aa ") == "SW1A 2AA"
    # Return original if not fixble
    assert fix(" 1A2aa ") == " 1A2aa "

    ## Outward code
    # Fixes LN format
    assert fix("01 OAA") == "O1 0AA"
    assert fix("SO OAA") == "S0 0AA"

    # Fixes L?? format
    assert fix("0W1 OAA") == "OW1 0AA"
    # ambiguous
    assert fix("S01 OAA") =="S01 0AA"
    assert fix("SO1 OAA") =="SO1 0AA"
    assert fix("SWO OAA") =="SWO 0AA"
    assert fix("SW0 OAA") == "SW0 0AA"

    # Fixes LNN?
    assert fix("0W1A OAA") == "OW1A 0AA" 
    assert fix("S01A OAA") == "SO1A 0AA"
    assert fix("SWOA OAA") == "SW0A 0AA"
    # ambiguous
    assert fix("SW10 OAA") == "SW10 0AA" 
    assert fix("SW1O OAA") == "SW1O 0AA" 

    ## Inward code
    # first character
    assert fix(" SW1A OAA") == "SW1A 0AA"
    # second character
    assert fix("SW1A 20A") == "SW1A 2OA"
    # third character
    assert fix("SW1A 2A0") == "SW1A 2AO"
    # 1 <=> I
    assert fix("SWIA 2AA") == "SW1A 2AA" 
    assert fix("1W1A 2AA") == "IW1A 2AA" 

def test_parsing():
    corpus = "this is a check to see if we can get post codes liek thia ec1r   1ub , and that e3 4ss. But also eh16 50y and ei412"          
    lst = parse_from_corpus(corpus)
    assert lst ==[
        Postcode(original='ec1r   1ub', postcode='EC1R 1UB', incode='1UB', outcode='EC1R', area='EC', district='EC1', sub_district='EC1R', sector='EC1R 1', unit='UB'), 
        Postcode(original='e3 4ss', postcode='E3 4SS', incode='4SS', outcode='E3', area='E', district='E3', sub_district=None, sector='E3 4', unit='SS')
        ]
    
    corpus = "this is a check to see if we can get post codes liek thia ec1r 1ub , and that e34ss. But also eh16 50y and ei412"          
    lst = parse_from_corpus(corpus, attempt_fix=True)
    assert lst ==[
        Postcode(original='ec1r 1ub', postcode='EC1R 1UB', incode='1UB', outcode='EC1R', area='EC', district='EC1', sub_district='EC1R', sector='EC1R 1', unit='UB'), 
        Postcode(original='e34ss', postcode='E3 4SS', incode='4SS', outcode='E3', area='E', district='E3', sub_district=None, sector='E3 4', unit='SS'), 
        Postcode(original='eh16 50y', postcode='EH16 5OY', incode='5OY', outcode='EH16', area='EH', district='EH16', sub_district=None, sector='EH16 5', unit='OY')
        ]

test_fix()
test_parsing()


import re
import logging
from dataclasses import dataclass
from typing import Union, List

from uk_postcodes_parsing.postcode_utils import *
from uk_postcodes_parsing.fix import fix

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('ukpostcode')

# Test for a valid postcode embedded in text
POSTCODE_CORPUS_REGEX = re.compile(r'[a-z]{1,2}\d[a-z\d]?\s*\d[a-z]{2}', re.I)
FIXABLE_POSTCODE_CORPUS_REGEX = re.compile("[a-z01]{1,2}[0-9oi][a-z\d]?\s*[0-9oi][a-z01]{2}")

def _parse(postcode: str) -> dict:
    if not is_valid(postcode):
        return None
    return {
        "postcode":  to_normalised(postcode),
        "incode":  to_incode(postcode),
        "outcode" :  to_outcode(postcode),
        "area":  to_area(postcode),
        "district":  to_district(postcode),
        "sub_district":  to_sub_district(postcode),
        "sector":  to_sector(postcode),
        "unit":  to_unit(postcode),
    }

def parse(postcode, attempt_fix=True):
    if is_valid(postcode):
        return Postcode(**_parse(postcode), original=postcode)
    if attempt_fix:
        fixed = fix(postcode)
        if is_valid(fixed):
            logger.info("Postcode Fixed: '%s' => '%s'", postcode, fixed)
            return Postcode(**_parse(fixed), original=postcode)
        logger.error("Unable to fix postcode")
    logger.error("Failed to parse postcode")
    return None

def parse_from_corpus(text: str, attempt_fix=False) -> List[str]:
    if attempt_fix:
         postcodes = re.findall(FIXABLE_POSTCODE_CORPUS_REGEX, text)
    else:
        postcodes = re.findall(POSTCODE_CORPUS_REGEX, text)
    logger.info("Found %d postcodes in corpus", len(postcodes))
    return list(map(parse, postcodes))

@dataclass
class Postcode():
    original: str
    postcode: str
    incode: str
    outcode: str
    area: str
    district: str
    sub_district: Union[str,None]
    sector: str
    unit: str


import re

FIXABLE_REGEX = re.compile(r"^\s*[a-z01]{1,2}[0-9oi][a-z\d]?\s*[0-9oi][a-z01]{2}\s*$", re.I)

def fix(s: str) -> str:
    match = FIXABLE_REGEX.match(s)
    if match is None:
        return s
    s = s.upper().strip().replace(r"\s+", "")
    inward = s[-3:].strip()
    outward = s[:-3].strip()
    return f"{coerce_outcode(outward)} {coerce_incode(inward)}"

def to_letter(c: str) -> str:
    return {
        "0": "O",
        "1": "I",
    }.get(c, c)

def to_number(c: str) -> str:
    return {
        "O": "0",
        "I": "1",
    }.get(c, c)


def coerce(pattern: str, input: str) -> str:
    """
    Coerce a given input into a pattern. For e.g. 
            pattern: "LN" => signifies a letter followed by a number
            input: "01" => changes to "O1". Where the first 0 should be a letter so it 
                is coverted from 0 to O. The Second 1 is already correct so no change.
    """
    return "".join(
            to_number(c) if target == "N" else to_letter(c) if target == "L" else c
            for c, target in zip(input, pattern)
    )

def coerce_outcode(i: str) -> str:
    """Coerce outcode based on length of outcode"""
    if len(i) == 2:
        return coerce("LN", i)
    elif len(i) == 3:
        return coerce("L??", i)
    elif len(i) == 4:
        return coerce("LLN?", i)
    else:
        return i

def coerce_incode(i :str) -> str:
    """Coerce incode"""
    return coerce('NLL', i)


import re
from typing import Union

# Tests for district 
DISTRICT_SPLIT_REGEX = re.compile(r'^([a-z]{1,2}\d)([a-z])$', re.I)

# Tests for the unit section of a postcode
UNIT_REGEX = re.compile(r'[a-z]{2}$', re.I)

# Tests for the inward code section of a postcode
INCODE_REGEX = re.compile(r'\d[a-z]{2}$', re.I)

# Tests for the outward code section of a postcode
OUTCODE_REGEX = re.compile(r'^[a-z]{1,2}\d[a-z\d]?$', re.I)

# Tests for a valid postcode
POSTCODE_REGEX = re.compile(r'^[a-z]{1,2}\d[a-z\d]?\s*\d[a-z]{2}$', re.I)

# Tests for the area section of a postcode
AREA_REGEX = re.compile(r'^[a-z]{1,2}', re.I)

def sanitize(s: str) -> str:
    return s.replace(' ', '').upper()

def is_valid(postcode: str) -> bool:
    return re.match(POSTCODE_REGEX, postcode) is not None

def is_valid_outcode(outcode: str) -> bool:
    return re.match(OUTCODE_REGEX, outcode) is not None

def to_normalised(postcode: str) -> Union[str, None]:
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    incode = to_incode(postcode)
    return None if incode is None else f'{outcode} {incode}'


def to_outcode(postcode: str) -> Union[str, None]:
    if not is_valid(postcode):
        return None
    return re.sub(INCODE_REGEX, '', sanitize(postcode))

def to_incode(postcode: str) -> Union[str, None]:
    if not is_valid(postcode):
        return
    incode = re.findall(INCODE_REGEX, sanitize(postcode))
    return incode[0] if incode else None

def to_area(postcode: str) -> Union[str, None]:
    """
    Returns a correctly formatted area given a postcode

    Args:
        postcode: The postcode to get the area of

    Returns:
        The area, or `None` if the postcode is invalid
    """

    if not is_valid(postcode):
        return None
    area = re.findall(AREA_REGEX, sanitize(postcode))
    return area[0] if area else None

def to_sector(postcode: str) -> Union[str, None]:
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    incode = to_incode(postcode)
    return None if incode is None else f'{outcode} {incode[0]}'


def to_unit(postcode: str) -> Union[str, None]:
    if not is_valid(postcode):
        return None
    unit = re.findall(UNIT_REGEX, sanitize(postcode))
    return unit[0] if unit else None

def to_district(postcode):
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    district = re.match(DISTRICT_SPLIT_REGEX, outcode)
    return district[1] if district else outcode

def to_sub_district(postcode):
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    split = re.match(DISTRICT_SPLIT_REGEX, outcode)
    return None if split is None else outcode